__author__ = 'Bohdan Mushkevych'

from threading import Lock
from subprocess import PIPE

import psutil
from psutil import TimeoutExpired

from launch import get_python, PROJECT_ROOT, PROCESS_STARTER
from synergy.conf import settings
from synergy.db.dao.box_configuration_dao import BoxConfigurationDao, QUERY_PROCESSES_FOR_BOX_ID
from synergy.supervisor.supervisor_constants import TRIGGER_INTERVAL
from synergy.supervisor.supervisor_configurator import get_box_id
from synergy.system.utils import remove_pid_file
from synergy.system.decorator import thread_safe
from synergy.system.repeat_timer import RepeatTimer
from synergy.system.synergy_process import SynergyProcess


class Supervisor(SynergyProcess):
    def __init__(self, process_name):
        super(Supervisor, self).__init__(process_name)
        self.thread_handlers = dict()
        self.lock = Lock()
        self.box_id = get_box_id(self.logger)
        self.bc_dao = BoxConfigurationDao(self.logger)
        self.logger.info('Started {0} with configuration for BOX_ID={1}'.format(self.process_name, self.box_id))

    def __del__(self):
        self.logger.info('Shutting down Supervisor...')
        for handler in self.thread_handlers:
            handler.cancel()
        self.thread_handlers.clear()
        super(Supervisor, self).__del__()

    # **************** Supervisor Methods ************************
    def _kill_process(self, box_config):
        """ method is called to kill a running process """
        try:
            self.logger.info('kill: {0} {{'.format(box_config.process_name))
            self.logger.info('target process pid={0}'.format(box_config.pid))
            if box_config.pid and psutil.pid_exists(box_config.pid):
                p = psutil.Process(box_config.pid)
                p.kill()
                p.wait()
                box_config.pid = None
                self.bc_dao.update(box_config)
                remove_pid_file(box_config.process_name)
        except Exception:
            self.logger.error('Exception on killing: {0}'.format(box_config.process_name), exc_info=True)
        finally:
            self.logger.info('}')

    def _start_process(self, box_config):
        if not self.bc_dao.ds.is_alive():
            # ping DB to make sure it is alive.
            # otherwise, processes will be spawned uncontrollably
            raise UserWarning('DB Down Exception: unable to reach db {0}'.format(self.bc_dao.ds))

        try:
            self.logger.info('start: {0} {{'.format(box_config.process_name))
            p = psutil.Popen([get_python(), PROJECT_ROOT + '/' + PROCESS_STARTER, box_config.process_name],
                             close_fds=True,
                             cwd=settings.settings['process_cwd'],
                             stdin=PIPE,
                             stdout=PIPE,
                             stderr=PIPE)
            box_config.pid = p.pid
            self.logger.info('Started {0} with pid = {1}'.format(box_config.process_name, p.pid))
        except Exception:
            box_config.set_process_pid(box_config.process_name, None)
            self.logger.error('Exception on starting: {0}'.format(box_config.process_name), exc_info=True)
        finally:
            self.bc_dao.update(box_config)
            self.logger.info('}')

    def _poll_process(self, box_config):
        """ between killing a process and its actual termination lies poorly documented requirement -
            <purging process' io pipes and reading exit status>.
            this can be done either by os.wait() or process.wait() """
        try:
            p = psutil.Process(box_config.pid)

            return_code = p.wait(timeout=0.01)
            if return_code is None:
                # process is already terminated
                self.logger.info('Process {0} is terminated'.format(box_config.process_name))
                return
            else:
                # process is terminated; possibly by OS
                box_config.pid = None
                self.bc_dao.update(box_config)
                self.logger.info('Process {0} got terminated. Cleaning up'.format(box_config.process_name))
        except TimeoutExpired:
            # process is alive and OK
            pass
        except Exception:
            self.logger.error('Exception on polling: {0}'.format(box_config.process_name), exc_info=True)

    def start(self, *_):
        """ reading box configurations and starting timers to start/monitor/kill processes """
        try:
            box_configurations = self.bc_dao.run_query(QUERY_PROCESSES_FOR_BOX_ID(self.box_id))

            for box_config in box_configurations:
                handler = RepeatTimer(TRIGGER_INTERVAL, self.manage_process, args=[box_config.process_name])
                self.thread_handlers[box_config.process_name] = handler
                handler.start()
                self.logger.info('Started Supervisor Thread for {0}, triggering every {1} seconds'
                                 .format(box_config.process_name, TRIGGER_INTERVAL))
        except LookupError as e:
            self.logger.error('Supervisor failed to start because of: {0}'.format(e))

    @thread_safe
    def manage_process(self, *args):
        """ reads box configuration and start/kill processes. performs process monitoring """
        process_name = args[0]
        try:
            box_config = self.bc_dao.get_one([self.box_id, process_name])
            if not box_config.is_on:
                if box_config.pid is not None:
                    self._kill_process(box_config)
                return

            if not box_config.pid or not psutil.pid_exists(box_config.pid):
                self._start_process(box_config)
            elif box_config.pid and psutil.pid_exists(box_config.pid):
                self._poll_process(box_config)
        except Exception as e:
            self.logger.error('Exception: {0}'.format(e), exc_info=True)


if __name__ == '__main__':
    from synergy.supervisor.supervisor_constants import PROCESS_SUPERVISOR

    source = Supervisor(PROCESS_SUPERVISOR)
    source.start()



__author__ = 'Bohdan Mushkevych'

from odm.document import BaseDocument
from odm.fields import StringField, ObjectIdField, ListField, IntegerField

MAX_NUMBER_OF_EVENTS = 128
TIMEPERIOD = 'timeperiod'
PROCESS_NAME = 'process_name'
STATE = 'state'
RELATED_UNIT_OF_WORK = 'related_unit_of_work'
NUMBER_OF_FAILURES = 'number_of_failures'

# contains list of last MAX_NUMBER_OF_EVENTS job events, such as emission of the UOW
EVENT_LOG = 'event_log'

# given Job was _not_ processed by aggregator because of multiple errors/missing data
# this state allows to mute current Job abd allow other timeperiods/Jobs to be processed
# only manual "re-processing" can re-run the skipped Job
STATE_SKIPPED = 'state_skipped'

# given Job was successfully processed by an aggregator
# no further processing for this Job is performed
STATE_PROCESSED = 'state_processed'

# no processing was performed for this Job
# no further processing for this Job is performed
STATE_NOOP = 'state_noop'

# Scheduler assumes that all timeperiod data is in the database, and asks an aggregator to run a "final" aggregation
# Job will be marked as STATE_PROCESSED afterwards if the processing succeed
STATE_FINAL_RUN = 'state_final_run'

# Aggregator is asked to perform a routine aggregation.
# Further state of the Job depends on the governing state machine:
# it could be either STATE_PROCESSED, STATE_IN_PROGRESS, STATE_NOOP, STATE_FINAL_RUN or STATE_SKIPPED
STATE_IN_PROGRESS = 'state_in_progress'

# Given timetable record serves as place-holder in the Tree
# TimeRecord can move to STATE_IN_PROGRESS
STATE_EMBRYO = 'state_embryo'


class Job(BaseDocument):
    """ class presents status for the time-period, and indicates whether data was process by particular process"""

    db_id = ObjectIdField('_id', null=True)
    process_name = StringField(PROCESS_NAME)
    timeperiod = StringField(TIMEPERIOD)
    state = StringField(STATE, choices=[STATE_IN_PROGRESS, STATE_PROCESSED, STATE_FINAL_RUN,
                                        STATE_EMBRYO, STATE_SKIPPED, STATE_NOOP])
    related_unit_of_work = ObjectIdField(RELATED_UNIT_OF_WORK)
    event_log = ListField(EVENT_LOG)
    number_of_failures = IntegerField(NUMBER_OF_FAILURES, default=0)

    @BaseDocument.key.getter
    def key(self):
        return self.process_name, self.timeperiod

    @key.setter
    def key(self, value):
        """ :param value: tuple (name of the process, timeperiod as string in Synergy Data format) """
        self.process_name = value[0]
        self.timeperiod = value[1]

    @property
    def is_active(self):
        return self.state in [STATE_FINAL_RUN, STATE_IN_PROGRESS, STATE_EMBRYO]

    @property
    def is_finished(self):
        return self.state in [STATE_PROCESSED, STATE_SKIPPED, STATE_NOOP]

    @property
    def is_processed(self):
        return self.state == STATE_PROCESSED

    @property
    def is_noop(self):
        return self.state == STATE_NOOP

    @property
    def is_skipped(self):
        return self.state == STATE_SKIPPED

    @property
    def is_embryo(self):
        return self.state == STATE_EMBRYO

    @property
    def is_in_progress(self):
        return self.state == STATE_IN_PROGRESS

    @property
    def is_final_run(self):
        return self.state == STATE_FINAL_RUN



__author__ = 'Bohdan Mushkevych'

import functools

from werkzeug.wrappers import Request
from synergy.mx.utils import jinja_env


def valid_action_request(method):
    """ wraps method with verification for is_request_valid"""

    @functools.wraps(method)
    def _wrapper(self, *args, **kwargs):
        assert isinstance(self, BaseRequestHandler)
        if not self.is_request_valid:
            return self.reply_bad_request()

        try:
            return method(self, *args, **kwargs)
        except UserWarning as e:
            return self.reply_server_error(e)
        except Exception as e:
            return self.reply_server_error(e)

    return _wrapper


def safe_json_response(method):
    """ makes sure the response' document has all leaf-fields converted to string """

    def _safe_document(document):
        """ function modifies the document in place
            it iterates over the json document and stringify all non-string types
            :return: modified document """
        assert isinstance(document, dict), 'Error: provided document is not of DICT type: {0}' \
            .format(document.__class__.__name__)

        for key, value in document.items():
            if isinstance(value, dict):
                document[key] = {k: str(v) for k, v in value.items()}
            elif isinstance(value, list):
                document[key] = [str(v) for v in value]
            else:
                document[key] = str(document[key])
        return document

    @functools.wraps(method)
    def _wrapper(self, *args, **kwargs):
        try:
            document = method(self, *args, **kwargs)
            return _safe_document(document)
        except Exception as e:
            return self.reply_server_error(e)

    return _wrapper


class BaseRequestHandler(object):
    def __init__(self, request, **values):
        assert isinstance(request, Request)
        self.scheduler = jinja_env.globals['mbean']
        self.logger = self.scheduler.logger
        self.request = request
        self.values = values
        self.request_arguments = request.args if request.args else request.form
        self.is_request_valid = False

    def reply_ok(self):
        return {'status': 'OK'}

    def reply_bad_request(self):
        self.logger.error('MX Bad Request: {0}'.format(self.request))
        return {}

    def reply_server_error(self, e):
        self.logger.error('MX Processing Exception: {0}'.format(e), exc_info=True)
        return {'status': 'Server Internal Error'}



__author__ = 'Bohdan Mushkevych'

import calendar
import collections

from synergy.system import time_helper
from synergy.system.time_qualifier import *


class TimeperiodDict(collections.MutableMapping):
    """ module represents a _smart_ dictionary, where key is a timeperiod
    TimeperiodDict allows timeperiod grouping, such as *every 3 hours* or *quarterly* or *weekly* """

    def __init__(self, time_qualifier, time_grouping, *args, **kwargs):
        assert time_qualifier in [QUALIFIER_HOURLY, QUALIFIER_DAILY, QUALIFIER_MONTHLY, QUALIFIER_YEARLY], \
            'time qualifier {0} is not supported by TimeperiodDict'.format(time_qualifier)
        super(TimeperiodDict, self).__init__()

        self.time_grouping = time_grouping
        self.time_qualifier = time_qualifier

        # validation section
        upper_boundary = self._get_stem_upper_boundary()
        assert 1 <= time_grouping <= upper_boundary

        # format: {grouped_timeperiod: value}
        self.data = dict()
        self.update(dict(*args, **kwargs))

    def _get_stem_upper_boundary(self, timeperiod=None):
        """
        :param timeperiod: optional parameter, applicable for QUALIFIER_DAILY qualifier only
        :return: upper boundary for dictionary's time_qualifier
        """

        if self.time_qualifier == QUALIFIER_HOURLY:
            upper_boundary = 23
        elif self.time_qualifier == QUALIFIER_DAILY:
            if timeperiod:
                # DAILY upper boundary is month-dependent
                # i.e. it is 28 for Feb 2015; and 31 for Mar 2015
                year, month, day, hour = time_helper.tokenize_timeperiod(timeperiod)
                monthrange_tuple = calendar.monthrange(int(year), int(month))
                upper_boundary = monthrange_tuple[1]
            else:
                upper_boundary = 28
        elif self.time_qualifier == QUALIFIER_MONTHLY:
            upper_boundary = 12
        elif self.time_qualifier == QUALIFIER_YEARLY:
            upper_boundary = 1
        else:
            raise ValueError('unknown time qualifier: {0}'.format(self.time_qualifier))
        return upper_boundary

    def _do_stem_grouping(self, timeperiod, stem):
        """
        method performs *timeperiod's stem grouping*.
        :param timeperiod: timeperiod to augment
        :param stem: inclusive lower boundary for the timeperiod's corresponding token. For instance:
                - for 2015010520 and QUALIFIER_MONTHLY, stem would be 01
                - for 2015010520 and QUALIFIER_DAILY, stem would be 05
                - for 2015010520 and QUALIFIER_HOURLY, stem would be 20
        :return: grouped stem. For instance:
                - for 2015010520 and QUALIFIER_MONTHLY and time_grouping=3, stem would be 03
                - for 2015010520 and QUALIFIER_DAILY and time_grouping=2, stem would be 06
                - for 2015010520 and QUALIFIER_HOURLY and time_grouping=8, stem would be 23
        """

        # exclude 00 from lower boundary, unless the grouping == 1
        lower_boundary = 0 if self.time_grouping == 1 else 1
        upper_boundary = self._get_stem_upper_boundary(timeperiod)

        for i in range(lower_boundary, upper_boundary):
            candidate = i * self.time_grouping
            if stem <= candidate <= upper_boundary:
                return candidate
        return upper_boundary

    def _translate_timeperiod(self, timeperiod):
        """ method translates given timeperiod to the grouped timeperiod """
        if self.time_grouping == 1:
            # no translation is performed for identity grouping
            return timeperiod

        # step 1: tokenize timeperiod into: (year, month, day, hour)
        # for instance: daily 2015031400 -> ('2015', '03', '14', '00')
        year, month, day, hour = time_helper.tokenize_timeperiod(timeperiod)

        # step 2: perform grouping on the stem
        # ex1: stem of 14 with grouping 20 -> 20
        # ex2: stem of 21 with grouping 20 -> 23
        if self.time_qualifier == QUALIFIER_HOURLY:
            stem = self._do_stem_grouping(timeperiod, int(hour))
            result = '{0}{1}{2}{3:02d}'.format(year, month, day, stem)
        elif self.time_qualifier == QUALIFIER_DAILY:
            stem = self._do_stem_grouping(timeperiod, int(day))
            result = '{0}{1}{2:02d}{3}'.format(year, month, stem, hour)
        else:  # self.time_qualifier == QUALIFIER_MONTHLY:
            stem = self._do_stem_grouping(timeperiod, int(month))
            result = '{0}{1:02d}{2}{3}'.format(year, stem, day, hour)
        return result

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        grouped_timeperiod = self._translate_timeperiod(key)
        return self.data.__getitem__(grouped_timeperiod)

    def __setitem__(self, key, value):
        grouped_timeperiod = self._translate_timeperiod(key)
        self.data.__setitem__(grouped_timeperiod, value)

    def __delitem__(self, key):
        self.data.__delitem__(key)

    def __iter__(self):
        return iter(self.data)

    def __contains__(self, key):
        grouped_timeperiod = self._translate_timeperiod(key)
        return self.data.__contains__(grouped_timeperiod)



__author__ = 'Bohdan Mushkevych'

from threading import Lock

from synergy.scheduler.scheduler_constants import QUEUE_UOW_STATUS, QUEUE_JOB_STATUS
from synergy.db.model.mq_transmission import MqTransmission
from synergy.mq.flopsy import PublishersPool
from synergy.system.decorator import thread_safe


class MqTransmitter(object):
    """ a class hosting several Message Queue helper methods to send MqTransmission """

    def __init__(self, logger):
        self.logger = logger
        self.lock = Lock()
        self.publishers = PublishersPool(self.logger)

    def __del__(self):
        try:
            self.logger.info('Closing Flopsy Publishers Pool...')
            self.publishers.close()
        except Exception as e:
            self.logger.error('Exception caught while closing Flopsy Publishers Pool: {0}'.format(e))

    @thread_safe
    def publish_managed_uow(self, uow):
        mq_request = MqTransmission(process_name=uow.process_name, record_db_id=uow.db_id)

        publisher = self.publishers.get(uow.process_name)
        publisher.publish(mq_request.document)
        publisher.release()

    @thread_safe
    def publish_freerun_uow(self, freerun_entry, uow):
        mq_request = MqTransmission(process_name=freerun_entry.process_name,
                                    entry_name=freerun_entry.entry_name,
                                    record_db_id=uow.db_id)

        publisher = self.publishers.get(freerun_entry.process_name)
        publisher.publish(mq_request.document)
        publisher.release()

    @thread_safe
    def publish_job_status(self, job_record, finished_only=True):
        if finished_only and not job_record.is_finished:
            return

        mq_request = MqTransmission(process_name=job_record.process_name, record_db_id=job_record.db_id)
        publisher = self.publishers.get(QUEUE_JOB_STATUS)
        publisher.publish(mq_request.document)
        publisher.release()

    @thread_safe
    def publish_uow_status(self, uow):
        mq_request = MqTransmission(process_name=uow.process_name, record_db_id=uow.db_id)

        publisher = self.publishers.get(QUEUE_UOW_STATUS)
        publisher.publish(mq_request.document)
        publisher.release()



__author__ = 'Bohdan Mushkevych'

QUALIFIER_REAL_TIME = '_real_time'
QUALIFIER_BY_SCHEDULE = '_by_schedule'
QUALIFIER_HOURLY = '_hourly'
QUALIFIER_DAILY = '_daily'
QUALIFIER_MONTHLY = '_monthly'
QUALIFIER_YEARLY = '_yearly'


# following dict is used by ProcessHierarchy
# for ordering ProcessEntries based on their time_qualifier
QUALIFIER_DICT = {
    QUALIFIER_YEARLY: 1000,
    QUALIFIER_MONTHLY: 900,
    QUALIFIER_DAILY: 800,
    QUALIFIER_HOURLY: 700
}



__author__ = 'Bohdan Mushkevych'

from werkzeug.utils import cached_property

from synergy.conf import settings
from synergy.system import time_helper
from synergy.conf import context
from synergy.mx.rest_model import RestTimetableTreeNode, RestJob
from synergy.mx.base_request_handler import BaseRequestHandler, valid_action_request


class TreeNodeDetails(BaseRequestHandler):
    def __init__(self, request, **values):
        super(TreeNodeDetails, self).__init__(request, **values)
        self.process_name = request.args.get('process_name')
        self.timeperiod = request.args.get('timeperiod')
        self.tree = self.scheduler.timetable.get_tree(self.process_name)
        self.is_request_valid = True if self.tree else False

    @classmethod
    def get_details(cls, node, as_model=False):
        """method returns either RestJob instance or corresponding document, depending on the as_model argument """
        rest_job = RestJob(
            process_name=node.process_name,
            timeperiod=node.timeperiod,
            time_qualifier=node.time_qualifier,
            number_of_children=len(node.children),
            number_of_failures='NA' if not node.job_record else node.job_record.number_of_failures,
            state='NA' if not node.job_record else node.job_record.state,
            event_log=[] if not node.job_record else node.job_record.event_log)

        if as_model:
            return rest_job
        else:
            return rest_job.document

    @cached_property
    @valid_action_request
    def details(self):
        rest_node = RestTimetableTreeNode()

        if not self.timeperiod:
            # return list of yearly nodes OR leafs for linear tree
            # limit number of children to return, since a linear tree can holds thousands of nodes
            children_keys = list(self.tree.root.children)
            sorted_keys = sorted(children_keys, reverse=True)
            sorted_keys = sorted_keys[:settings.settings['mx_children_limit']]
            for key in sorted_keys:
                child = self.tree.root.children[key]
                rest_node.children[key] = TreeNodeDetails.get_details(child)

        else:
            time_qualifier = context.process_context[self.process_name].time_qualifier
            self.timeperiod = time_helper.cast_to_time_qualifier(time_qualifier, self.timeperiod)
            node = self.tree.get_node(self.process_name, self.timeperiod)
            rest_node.node = TreeNodeDetails.get_details(node, as_model=True)
            for key, child in node.children.items():
                rest_node.children[key] = TreeNodeDetails.get_details(child)

        return rest_node.document



__author__ = 'Bohdan Mushkevych'

from six import string_types
from datetime import datetime, timedelta

from synergy.system.repeat_timer import RepeatTimer

TIME_OF_DAY_FORMAT = "%H:%M"
EVERY_DAY = '*'        # marks every day as suitable to trigger the event
TRIGGER_INTERVAL = 30  # half a minute


class EventTime(object):
    def __init__(self, trigger_frequency):
        self.trigger_frequency = trigger_frequency

        tokens = self.trigger_frequency.split('-')
        if len(tokens) > 1:
            # Day of Week is provided
            self.day_of_week = tokens[0]
            self.time_of_day = datetime.strptime(tokens[1], TIME_OF_DAY_FORMAT)
        else:
            # Day of Week is not provided. Assume every day of the week
            self.day_of_week = EVERY_DAY
            self.time_of_day = datetime.strptime(tokens[0], TIME_OF_DAY_FORMAT)

    def __str__(self):
        return 'EventTime: day_of_week={0} time_of_day={1}'\
               .format(self.day_of_week, self.time_of_day.strftime(TIME_OF_DAY_FORMAT))

    def __repr__(self):
        return '{0}-{1}'.format(self.day_of_week, self.time_of_day.strftime(TIME_OF_DAY_FORMAT))

    def __eq__(self, other):
        if not isinstance(other, EventTime):
            return False

        return self.time_of_day == other.time_of_day \
            and (self.day_of_week == other.day_of_week
                 or self.day_of_week == EVERY_DAY
                 or other.day_of_week == EVERY_DAY)

    def __hash__(self):
        return hash((self.day_of_week, self.time_of_day))

    def next_trigger_frequency(self, utc_now=None):
        """ :param utc_now: optional parameter to be used by Unit Tests as a definition of "now"
            :return: datetime instance presenting next trigger time of the event """
        if utc_now is None:
            utc_now = datetime.utcnow()

        def wind_days(start_date):
            while True:
                if self.day_of_week == EVERY_DAY or start_date.weekday() == int(self.day_of_week):
                    return start_date.replace(hour=self.time_of_day.hour, minute=self.time_of_day.minute)
                else:
                    start_date += timedelta(days=1)

        if utc_now.time() > self.time_of_day.time():
            return wind_days(utc_now + timedelta(days=1))
        else:
            return wind_days(utc_now)

    @classmethod
    def utc_now(cls):
        utc_now = datetime.utcnow()
        return EventTime('{0}-{1}'.format(utc_now.weekday(), utc_now.strftime(TIME_OF_DAY_FORMAT)))


class EventClock(object):
    """ This class triggers on predefined time set in format 'day_of_week-HH:MM' or 'HH:MM'
    Maintaining API compatibility with the RepeatTimer class """

    def __init__(self, interval, call_back, args=None, kwargs=None):
        if not kwargs: kwargs = {}
        if not args: args = []

        self.timestamps = []
        self.change_interval(interval)

        self.args = args
        self.kwargs = kwargs
        self.call_back = call_back
        self.handler = RepeatTimer(TRIGGER_INTERVAL, self.manage_schedule)
        self.activation_dt = None

    def _trigger_now(self):
        if self.activation_dt is not None and datetime.utcnow() - self.activation_dt < timedelta(minutes=1):
            # the event was already triggered within 1 minute. no need to trigger it again
            return
        self.call_back(*self.args, **self.kwargs)
        self.activation_dt = datetime.utcnow()

    def manage_schedule(self, *_):
        current_time = EventTime.utc_now()
        if current_time in self.timestamps:
            self._trigger_now()

    def start(self):
        self.handler.start()

    def cancel(self):
        self.handler.cancel()

    def trigger(self):
        current_time = EventTime.utc_now()
        if current_time not in self.timestamps:
            self._trigger_now()
        else:
            # leave it to the regular flow to trigger the call_back via manage_schedule method
            pass

    def change_interval(self, value):
        """ :param value: list of strings in format 'Day_of_Week-HH:MM' """
        assert not isinstance(value, string_types)
        self.timestamps = []

        for timestamp in value:
            event = EventTime(timestamp)
            self.timestamps.append(event)

    def next_run_in(self, utc_now=None):
        """ :param utc_now: optional parameter to be used by Unit Tests as a definition of "now"
            :return: timedelta instance presenting amount of time before the trigger is triggered next time
         or None if the EventClock instance is not running """
        if utc_now is None:
            utc_now = datetime.utcnow()

        if self.is_alive():
            smallest_timedelta = timedelta(days=99, hours=0, minutes=0, seconds=0, microseconds=0, milliseconds=0)
            for event_time in self.timestamps:
                next_trigger = event_time.next_trigger_frequency(utc_now)
                if next_trigger - utc_now < smallest_timedelta:
                    smallest_timedelta = next_trigger - utc_now
            return smallest_timedelta

        else:
            return None

    def is_alive(self):
        return self.handler.is_alive()



__author__ = 'Bohdan Mushkevych'

from threading import RLock

from bson.objectid import ObjectId
from pymongo import ASCENDING
from pymongo.errors import DuplicateKeyError as MongoDuplicateKeyError

from synergy.conf import context
from synergy.db.error import DuplicateKeyError
from synergy.db.manager import ds_manager
from synergy.db.model import unit_of_work
from synergy.db.model.unit_of_work import UnitOfWork
from synergy.scheduler.scheduler_constants import COLLECTION_UNIT_OF_WORK
from synergy.system import time_helper
from synergy.system.decorator import thread_safe
from synergy.system.time_qualifier import *

QUERY_GET_FREERUN_SINCE = lambda timeperiod, include_running, include_processed, include_noop, include_failed: {
    unit_of_work.TIMEPERIOD: {'$gte': timeperiod},
    unit_of_work.UNIT_OF_WORK_TYPE: unit_of_work.TYPE_FREERUN,
    unit_of_work.STATE: {'$in': [unit_of_work.STATE_PROCESSED if include_processed else None,
                                 unit_of_work.STATE_IN_PROGRESS if include_running else None,
                                 unit_of_work.STATE_REQUESTED if include_running else None,
                                 unit_of_work.STATE_CANCELED if include_failed else None,
                                 unit_of_work.STATE_INVALID if include_failed else None,
                                 unit_of_work.STATE_NOOP if include_noop else None]}
}


class UnitOfWorkDao(object):
    """ Thread-safe Data Access Object from units_of_work table/collection """

    def __init__(self, logger):
        super(UnitOfWorkDao, self).__init__()
        self.logger = logger
        self.lock = RLock()
        self.ds = ds_manager.ds_factory(logger)

    @thread_safe
    def get_one(self, key):
        """ method finds unit_of_work record and returns it to the caller"""
        query = {'_id': ObjectId(key)}
        collection = self.ds.connection(COLLECTION_UNIT_OF_WORK)

        document = collection.find_one(query)
        if document is None:
            msg = 'UOW with ID={0} was not found'.format(key)
            self.logger.warning(msg)
            raise LookupError(msg)
        return UnitOfWork.from_json(document)

    @thread_safe
    def get_reprocessing_candidates(self, since=None):
        """ method queries Unit Of Work whose <start_timeperiod> is younger than <since>
        and who could be candidates for re-processing """
        collection = self.ds.connection(COLLECTION_UNIT_OF_WORK)

        query = {unit_of_work.STATE: {'$in': [unit_of_work.STATE_IN_PROGRESS,
                                              unit_of_work.STATE_INVALID,
                                              unit_of_work.STATE_REQUESTED]},
                 unit_of_work.UNIT_OF_WORK_TYPE: unit_of_work.TYPE_MANAGED}

        if since is None:
            cursor = collection.find(query).sort('_id', ASCENDING)
            candidates = [UnitOfWork.from_json(document) for document in cursor]
        else:
            candidates = []
            yearly_timeperiod = time_helper.cast_to_time_qualifier(QUALIFIER_YEARLY, since)
            query[unit_of_work.START_TIMEPERIOD] = {'$gte': yearly_timeperiod}

            cursor = collection.find(query).sort('_id', ASCENDING)
            for document in cursor:
                uow = UnitOfWork.from_json(document)
                if uow.process_name not in context.process_context:
                    # this is a decommissioned process
                    continue

                time_qualifier = context.process_context[uow.process_name].time_qualifier
                if time_qualifier == QUALIFIER_REAL_TIME:
                    time_qualifier = QUALIFIER_HOURLY
                process_specific_since = time_helper.cast_to_time_qualifier(time_qualifier, since)

                if process_specific_since <= uow.start_timeperiod:
                    candidates.append(uow)

        if len(candidates) == 0:
            raise LookupError('MongoDB has no UOW reprocessing candidates')
        return candidates

    @thread_safe
    def get_by_params(self, process_name, timeperiod, start_id, end_id):
        """ method finds unit_of_work record and returns it to the caller"""
        query = {unit_of_work.PROCESS_NAME: process_name,
                 unit_of_work.TIMEPERIOD: timeperiod,
                 unit_of_work.START_ID: start_id,
                 unit_of_work.END_ID: end_id}
        collection = self.ds.connection(COLLECTION_UNIT_OF_WORK)

        document = collection.find_one(query)
        if document is None:
            raise LookupError('UOW satisfying query {0} was not found'.format(query))
        return UnitOfWork.from_json(document)

    @thread_safe
    def update(self, instance):
        """ method finds unit_of_work record and change its status"""
        assert isinstance(instance, UnitOfWork)
        if instance.db_id:
            query = {'_id': ObjectId(instance.db_id)}
        else:
            query = {unit_of_work.PROCESS_NAME: instance.process_name,
                     unit_of_work.TIMEPERIOD: instance.timeperiod,
                     unit_of_work.START_ID: instance.start_id,
                     unit_of_work.END_ID: instance.end_id}
        self.ds.update(COLLECTION_UNIT_OF_WORK, query, instance)
        return instance.db_id

    @thread_safe
    def insert(self, instance):
        """ inserts a unit of work into MongoDB.
        :raises DuplicateKeyError: if such record already exist """
        assert isinstance(instance, UnitOfWork)
        collection = self.ds.connection(COLLECTION_UNIT_OF_WORK)
        try:
            return collection.insert_one(instance.document).inserted_id
        except MongoDuplicateKeyError as e:
            exc = DuplicateKeyError(instance.process_name,
                                    instance.start_timeperiod,
                                    instance.start_id,
                                    instance.end_id,
                                    e)
            raise exc

    @thread_safe
    def remove(self, uow_id):
        collection = self.ds.connection(COLLECTION_UNIT_OF_WORK)
        collection.delete_one({'_id': ObjectId(uow_id)})

    @thread_safe
    def run_query(self, query):
        """ method runs the query and returns a list of filtered UnitOfWork records """
        cursor = self.ds.filter(COLLECTION_UNIT_OF_WORK, query)
        return [UnitOfWork.from_json(document) for document in cursor]

    def recover_from_duplicatekeyerror(self, e):
        """ method tries to recover from DuplicateKeyError """
        if isinstance(e, DuplicateKeyError):
            try:
                return self.get_by_params(e.process_name, e.timeperiod, e.start_id, e.end_id)
            except LookupError as e:
                self.logger.error('Unable to recover from DuplicateKeyError error due to {0}'.format(e), exc_info=True)
        else:
            msg = 'Unable to recover from DuplicateKeyError due to unspecified UOW primary key'
            self.logger.error(msg)



__author__ = 'Bohdan Mushkevych'

TRIGGER_INTERVAL = 5    # seconds between checking process state

PROCESS_SUPERVISOR = 'Supervisor'
TOKEN_SUPERVISOR = 'supervisor'

COLLECTION_BOX_CONFIGURATION = 'box_configuration'
